#ifndef Corrade_Containers_Reference_h
#define Corrade_Containers_Reference_h
/*
    This file is part of Corrade.

    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
                2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
              Vladimír Vondruš <mosra@centrum.cz>

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

/* std::declval() is said to be in <utility> but libstdc++, libc++ and MSVC STL
   all have it directly in <type_traits> because it just makes sense */
#include <type_traits>

#ifndef CORRADE_SINGLES_NO_DEBUG
#include "Corrade/Utility/Debug.h"
#endif

/** @file
 * @brief Class @ref Corrade::Containers::Reference
 * @see @ref Corrade/Containers/ReferenceStl.h
 */

namespace Corrade { namespace Containers {

namespace Implementation {
    template<class, class> struct ReferenceConverter;
}

/**
@brief Lightweight non-owning reference wrapper

Equivalent to @ref std::reference_wrapper from C++11, provides a copyable
non-owning wrapper over l-value references to allow storing them in containers.
The @ref reference() convenience function provides an equivalent to
@ref std::ref(), see its documentation for a potential use case. An equivalent
to @ref std::cref() is not deemed necessary --- in most contexts where
@ref Reference is used, passing a plain reference works just as well, with no
need to explicitly wrap it first. This class is trivially copyable
(@ref std::reference_wrapper is guaranteed to be so only since C++17) and also
works on incomplete types, which @ref std::reference_wrapper knows since C++20.

This class is exclusively for l-value references. If you want to accept r-value
references instead, use a @ref MoveReference; if you want to accept both, use
an @ref AnyReference.

@section Containers-Reference-stl STL compatibility

Instances of @ref Reference are implicitly convertible to and from
@ref std::reference_wrapper if you include @ref Corrade/Containers/ReferenceStl.h.
The conversion is provided in a separate header to avoid unconditional
@cpp #include <functional> @ce, which significantly affects compile times.
Example:

@snippet Containers-stl.cpp Reference

<b></b>

@m_class{m-block m-success}

@par Single-header version
    This class is also available as a single-header, dependency-less
    [CorradeReference.h](https://github.com/mosra/magnum-singles/tree/master/CorradeReference.h)
    library in the Magnum Singles repository for easier integration into your
    projects. See @ref corrade-singles for more information. The above
    mentioned STL compatibility is included as well, but disabled by default.
    Enable it by specifying @cpp #define CORRADE_REFERENCE_STL_COMPATIBILITY @ce
    before including the file. Including it multiple times with different
    macros defined works as well.

@see @ref reference(T&), @ref Pointer, @ref Optional
*/
template<class T> class Reference {
    public:
        /**
         * @brief Value type
         * @m_since_latest
         */
        typedef T Type;

        /** @brief Constructor */
        constexpr /*implicit*/ Reference(T& reference) noexcept: _reference{&reference} {}

        /**
         * @brief Construct a reference from external representation
         *
         * @see @ref Containers-Reference-stl
         */
        template<class U, class = decltype(Implementation::ReferenceConverter<T, U>::from(std::declval<U>()))> constexpr /*implicit*/ Reference(U other) noexcept: Reference{Implementation::ReferenceConverter<T, U>::from(other)} {}

        /**
         * @brief Construction from r-value references is not allowed
         *
         * A @ref MoveReference can be created from r-value references instead.
         * @todo Fix LWG 2993 / LWG 3041 (details in the skipped test)
         */
        Reference(T&&) = delete;

        /**
         * @brief Construct a reference from another of a derived type
         *
         * Expects that @p T is a base of @p U.
         */
        template<class U, class = typename std::enable_if<std::is_base_of<T, U>::value>::type> constexpr /*implicit*/ Reference(Reference<U> other) noexcept: _reference{other._reference} {}

        /**
         * @brief Convert the reference to external representation
         *
         * @see @ref Containers-Reference-stl
         */
        template<class U, class = decltype(Implementation::ReferenceConverter<T, U>::to(std::declval<Reference<T>>()))> constexpr /*implicit*/ operator U() const {
            return Implementation::ReferenceConverter<T, U>::to(*this);
        }

        /** @brief Underlying reference */
        constexpr /*implicit*/ operator T&() const { return *_reference; }
        /** @overload */
        constexpr /*implicit*/ operator Reference<const T>() const { return *_reference; }

        /** @brief Underlying reference */
        constexpr T& get() const { return *_reference; }

        /**
         * @brief Access the underlying reference
         *
         * @ref get(), @ref operator*()
         */
        constexpr T* operator->() const { return _reference; }

        /**
         * @brief Access the underlying reference
         *
         * @see @ref get(), @ref operator->()
         */
        constexpr T& operator*() const { return *_reference; }

    private:
        /* For the conversion constructor */
        template<class U> friend class Reference;

        T* _reference;
};

/** @relatesalso Reference
@brief Make a reference wrapper
@m_since_latest

Convenience alternative to @ref Reference::Reference(T&). Thw following two
lines are equivalent:

@snippet Containers.cpp reference

Useful for example when iterating a list of variables that need to be modified
in-place, and where the code would otherwise have to be extra verbose or would
fall back to raw pointers and risk them getting @cpp nullptr @ce:

@snippet Containers.cpp reference-iterating
*/
template<class T> constexpr Reference<T> reference(T& reference) {
    return Reference<T>{reference};
}

#ifndef CORRADE_SINGLES_NO_DEBUG
/** @debugoperator{Reference} */
template<class T> Utility::Debug& operator<<(Utility::Debug& debug, Reference<T> value) {
    return debug << value.get();
}
#endif

}}

#endif
